'''
Demonstrate an n-way set-associative cache with adjustable
blocksize. Selecting a row in memory highlights the appropriate
data word in the cache and sets the bits in the address
fields at the top. Pushing the read button either gets a
hit and fills in the data field at the top or beeps and
fills in the data row of the cache.

An n-way set-associative cache has n blocks per set. In the display
we use, the blocks are divided into groups that are separated
horizontally. Each row in the entire structure is a set.
'''
import tkinter
from tkinter import ttk
import jtk
import random
import math

##some functions we will need to calculate critical sizes
def blocks(cacheSize, blockSize):
    return cacheSize // blockSize

def sets(blocks, blocksPerSet):
    return blocks // blocksPerSet

def memoryBits(memorySize):
    return int(math.log(memorySize, 2))

def indexBits(sets):
    return int(math.log(sets, 2))

def blockBits(blockSize):
    retval =  int(math.log(blockSize, 2))
    return retval

def tagBits(memoryBits, indexBits, blockBits):
    return memoryBits - indexBits - blockBits

def colors(blockSize, sets):                    ## how many colors do we need?
    return blockSize * sets // 4                ## color every fourth memory entry

def powerOfTwo(n):
    ln = math.log(n, 2)
    return n >= 1 and ln == int(ln)


#################### tune these four integers to try out different cache organizations ############
MEMORYSIZE = 1024       ### words, NOT bytes
CACHESIZE = 128          ### words, NOT bytes
BLOCKSIZE = 4           ### words, NOT bytes
ASSOCIATIVITY = 2       ### blocks per set
#################### do NOT change anything below this line!!!!!!!!!!!
## Now do some sanity checks,
assert powerOfTwo(MEMORYSIZE) , "MEMORYSIZE must be a power of two"
assert powerOfTwo(CACHESIZE) , "CACHESIZE must be a power of two"
assert powerOfTwo(BLOCKSIZE) , "BLOCKSIZE must be a power of two"
assert powerOfTwo(ASSOCIATIVITY) , "ASSOCIATIVITY must be a power of two"
assert MEMORYSIZE > CACHESIZE >= BLOCKSIZE * ASSOCIATIVITY , "WORDS per SET must not be greater than CACHESIZE"

## and calculate some more numbers we will need.
BLOCKS = CACHESIZE // BLOCKSIZE
SETS = BLOCKS // ASSOCIATIVITY             ## = the nunber of sets; blocks per set = associativity
MEMORYBITS = int(math.log(MEMORYSIZE, 2))
INDEXBITS = int(math.log(SETS, 2))
BLOCKBITS = int(math.log(BLOCKSIZE, 2))
TAGBITS = MEMORYBITS - INDEXBITS - BLOCKBITS


class AddrPanel(ttk.Frame): 
    def __init__(self, parent, indexBits, blockBits, tagBits):
        super().__init__(parent)
        self.tag = AddrField(self, 2, tagBits)
        self.index = AddrField(self, 2, indexBits)
        self.offset = AddrField(self, 2, blockBits)
        self.byte = AddrField(self, 2, 2)  ## byte bits is always 2
        rb = ttk.Button(self, text="read", command=self.master.read)
        self.data = AddrField(self, 10, 4)

        ttk.Label(self, text="tag").grid(row=0, column=0)        
        ttk.Label(self, text="idx").grid(row=0, column=1)        
        ttk.Label(self, text="off").grid(row=0, column=2)        
        ttk.Label(self, text="byte").grid(row=0, column=3)        
        ttk.Label(self, text="data").grid(row=0, column=5, padx=20)        
        self.tag.grid(row=1, column=0)
        self.index.grid(row=1, column=1)
        self.offset.grid(row=1, column=2)
        self.byte.grid(row=1, column=3)
        rb.grid(row=1, column=4, padx=10)
        self.data.grid(row=1, column=5, padx=20)
        self.byte.setFromString("00")             ## These zeroes are fixed and will not change.

    def setAddr(self, tagVal:int, idxVal:int, offVal:int):          ## parameters are integers
        self.tag.set(tagVal)
        self.index.set(idxVal)
        self.offset.set(offVal)

    def setData(self, dataVal:int):
        self.data.set(dataVal)

    def clearData(self):
        self.data.setFromString("")

    def getAddr(self):
        tag = self.tag.get()
        idx = self.index.get()
        off = self.offset.get()
        return tag, idx, off

class AddrField(tkinter.Entry):
    def __init__(self, parent, base, digits, **kwargs):
        super().__init__(parent, width=digits, state="readonly", readonlybackground="white", **kwargs)
        self.base = base
        self.digits = digits
        
    def setFromString(self, newValue:str):              ## newValue is string
        self.config(state="normal")
        self.delete(0, "end")
        self.insert(0, newValue)
        self.config(state="readonly")

    def set(self, newValue:int):                ## newValue and digits are integers
        val = jtk.radStr(newValue, self.base, self.digits)  ## val == "" if self.digits == 0
##        print("setting: {}".format(val))
        self.setFromString(val)

    def get(self) -> int:
        val = super().get()                     ## gets a binary numeral as a str
        if val is not "":                       ## returns an int if it can
            return int(val, self.base)
        else:
            return 0

class BlockGroup(ttk.Frame):
    def __init__(self, parent, blockSize, sets, indexBits, tagBits):
        super().__init__(parent)
        self.sets = sets
        headers = ["Idx", "V", "Tag"]
        for i in range(blockSize):
            headers.append("Dat" + str(i))
        self.cols = cols = len(headers)
        for i in range(cols):
            ttk.Label(self, text=headers[i]).grid(row=0, column=i)
        self.table = table = []
        for i in range(sets):
            r = []
            for j in range(cols):
                e = tkinter.Entry(self)
                e.grid(row=i+1, column=j)
                r.append(e)
            table.append(r)
        for i in range(sets):
            table[i][0].insert(0, jtk.radStr(i, 2, indexBits))                                  ## index     
            table[i][0].config(width=indexBits+1, state="readonly")
##            if i % 4 == 0:                                            ## may color the cache some day
##                colorNum = (i // 4) % COLORS
##                table[i][0].config(readonlybackground=jtk.DISTINCT_COLORS[colorNum])
            table[i][1].insert(0, 0)                                                            ## valid
            table[i][1].config(width=2, state="readonly")
            table[i][2].insert(0, "0"*tagBits)                                ## tag
            table[i][2].config(width=tagBits, state="readonly")
            for j in range(3, cols):                                                            ## tag and data
                table[i][j].insert(0, "000")
                table[i][j].config(width=4, state="readonly")
                
    ## apply get() to each element in table[row], gather the results into a list, and return the list
    def getRow(self, row):
        return [e.get() for e in self.table[row]]

    def setRow(self, index, contents):              ## copy items from contents into row chosen by index
        for i in range(len(contents)):
            self.setItem(index, i+1, contents[i])   ## leave column 0 alone; it contains the row index

    def setItem(self, row, column, newVal):
        self.table[row][column].config(state="normal")
        self.table[row][column].delete(0, 'end')
        self.table[row][column].insert(0, newVal)
        self.table[row][column].config(state="readonly")
    
    def highlight(self, idx, off):
        for i in range(self.sets):
            for j in range(3, self.cols):
                if i == idx and j == 3 + off:
                    self.table[i][j].config(readonlybackground="pink")
                else:
                    self.table[i][j].config(readonlybackground="white")

    def find(self, tagVal, idxVal, offVal) -> (bool, int) :
        theRow = self.getRow(idxVal)
        cvalid = int(theRow[1], 2)
        ctag =   int(theRow[2], 2)
        if cvalid == 1 and ctag == tagVal:  ######### cache hit
            val = theRow[3 + offVal]        ## get correct value from the block
            print("hit got: ", val)
            return True, int(val)
        else:
            return False, 0

class Cache(ttk.Frame):                     ## needs ASSICIATIVITY
    def __init__(self, parent, associativity, blockSize, sets, indexBits, tagBits):
        super().__init__(parent)
        self.associativity = associativity
        self.blockGroups = []
        for i in range(associativity):
            self.blockGroups.append(BlockGroup(self, blockSize, sets, indexBits, tagBits))
            ttk.Label(self, text="Way {}".format(i)).grid(row=0, column=i)
            self.blockGroups[i].grid(row=1, column=i, padx=10)

    def highlight(self, idx, off):
        for i in range(self.associativity):
            self.blockGroups[i].highlight(idx, off)
        
    def find(self, tagVal, idxVal, offVal):
        for i in range(self.associativity):                                  ## search all blockGroups
            found, val = self.blockGroups[i].find(tagVal, idxVal, offVal)
            if found:
                return True, val
        return False, 0                                                 ## not found anywhere

    def setRow(self, index, contents):                             ## copy items from contents into row chosen by index
        whichBlock = random.randint(0, self.associativity - 1)    ## randomly choose a block in the set
        self.blockGroups[whichBlock].setRow(index, contents)      ## randint includes both endpoints
        ## blink label of blockGroups[whichBlock]

class Memory(jtk.McListBox):
    def __init__(self, parent, headers, memorySize, colors):
        super().__init__(parent, headers, selectmode="browse")
        for addr in range(memorySize):
            self.tree.insert('', 'end', self.idFromIndex(addr), values=(4*addr, random.randint(0, 1023)))
            if addr % 4 == 0 and colors > 0:
                colorNum = (addr // 4) % colors
                color = jtk.DISTINCT_COLORS[colorNum]
                id = self.idFromIndex(addr)
                self.tree.item(id, tags=color)
                self.tree.tag_configure(color, background=color)

        self.tree["height"] = 35
        self.tree.selection_set("I000")  ## select first row
        self.tree.bind("<<TreeviewSelect>>", self.selected)
        self.verticalScroll()

    def selected(self, evt=None):                           ## called anytime selection occurs
        memaddr, memval = self.getSelectionValues()         ## get address of selected memory location
        memaddr = str(int(memaddr) // 4)                    ## divide by four since we are sending it out
        self.master.setAddr(memaddr)          

    def getValue(self, index):                      ## get value at a memory location passed in from outside
        addr, value = self.getByIndex(index)        ## outside address is really an index, so we leave it alone
        return value                                ## will be a String

class CacheApp(ttk.Frame):
    def __init__(self, parent, memorySize, cacheSize, blockSize, associativity):
        super().__init__(parent)
        self.memorySize = memorySize
        self.cacheSize = cacheSize
        self.blockSize = blockSize
        self.associativity = associativity
        self.blocks = cacheSize // blockSize
        self.sets = self.blocks // associativity
        self.memoryBits = int(math.log(memorySize, 2))
        self.indexBits = int(math.log(self.sets, 2))
        self.blockBits = int(math.log(blockSize, 2))
        self.tagBits = self.memoryBits - self.indexBits - self.blockBits
        colors = self.sets * blockSize // 4              ## color every fourth memory entry
        assert colors <= len(jtk.DISTINCT_COLORS) , "not enough colors for your CACHESIZE; try reducing it."
##        assert(self.memorySize == MEMORYSIZE)              ### these were all for debugging
##        assert(self.cacheSize == CACHESIZE)
##        assert(self.blockSize == BLOCKSIZE)
##        assert(self.associativity == ASSOCIATIVITY)
##        assert(self.blocks == BLOCKS)
##        assert(self.sets == SETS)
##        assert(self.memoryBits == MEMORYBITS)
##        assert(self.indexBits == INDEXBITS)
##        assert(self.blockBits == BLOCKBITS)
##        assert(self.tagBits == TAGBITS)
        
        ass = ""
        if self.sets == 1 or associativity == 1:
            if associativity == 1:
                ass = ass + "Direct-Mapped "
            if self.sets == 1:
                ass = ass + "Fully-Associative "
        else:
            ass = "{}-Way Set-Associative ".format(associativity)
        parent.title("{}-Byte {} Cache with Blocksize {}".format(cacheSize*4, ass, blockSize))

        print("{}  {} = {} * {} = {} * {} * {}".format(memorySize, cacheSize, blockSize, self.blocks,
                                                       self.sets, associativity, blockSize))
        print("{} = {} + {} + {}".format(self.memoryBits, self.tagBits, self.indexBits, self.blockBits))
        
        self.ap = AddrPanel(self, self.indexBits, self.blockBits, self.tagBits)
        self.cache = Cache(self, associativity, blockSize, self.sets, self.indexBits, self.tagBits)
        self.mem = Memory(self, ("address", "contents"), memorySize, colors)
        self.ap.pack(side="top", pady="10", padx="125")
        self.cache.pack(side="left", padx = 50)
        self.mem.pack(side="left", padx = 15)

    def splitAddr(self, addr):          ## compute tag, index, and block offset from an address
        addr = int(addr)                ## When we say "block offset" we mean the position,
        offset = addr % self.blockSize       ## i.e. offset, of the word's address within it's block.
        blockNum = addr // self.blockSize    ## addr is a word address, i.e. byte address // 4
        index = blockNum % self.sets
        tag = blockNum // self.sets        
        print("addresses:", addr, tag, index, offset, self.assembleAddr(tag, index, offset))
        assert self.assembleAddr(tag, index, offset) == addr, "splitAddr and assembleAddr are not inverses"
        return tag, index, offset

    def assembleAddr(self, tag, index, offset):     ## compute address from tag, index, and block offset
        return (tag * self.sets + index) * self.blockSize + offset

    def setAddr(self, memaddr):                     ## When an address is selected, show that in the Cache
        tag, index, offset = self.splitAddr(memaddr)        ## and in the AddressPanel.
        self.ap.setAddr(tag, index, offset)
        self.cache.highlight(index, offset)

    def read(self):                                  ## callback for pushes on the "Read" button
        tagVal, idxVal, offVal = self.ap.getAddr()  ## gets address in pieces
        print(tagVal, idxVal, offVal)
        address = self.assembleAddr(tagVal, idxVal, offVal)  ## assemble complete address from pieces
        hit, val = self.cache.find(tagVal, idxVal, offVal)
        if hit:
            self.ap.setData(val)
            if address + 1 < self.memorySize:    # advance to next address automatically; saves button pushes
                self.mem.selectByIndex(address + 1)
        else:                                       ######### cache miss
            self.ap.clearData()
            self.bell()
            print("miss; installing block ")
            firstAddressInBlock = address - offVal
            newRow = ["1", jtk.radStr(tagVal, 2, self.tagBits)]                 ## valid bit and  tag
            for i in range(self.blockSize):
                newRow.append(self.mem.getValue(firstAddressInBlock + i))       ## data values
            self.cache.setRow(idxVal, newRow)

if __name__  == '__main__':    
    root = tkinter.Tk()
    app = CacheApp(root, MEMORYSIZE, CACHESIZE, BLOCKSIZE, ASSOCIATIVITY)
    app.pack()
    root.mainloop()
