"""
Organize the buttons and modulus altering stuff into a class.
"""
import tkinter
from tkinter import ttk
import jtk
import logging
import model
import math
import time

class ButtonPanel(ttk.Frame):
    def __init__(self, parent, theModel):
        super().__init__(parent)
        self.theModel = theModel
        upButton = ttk.Button(self, text='+1', command=self.doUp)
        downButton = ttk.Button(self, text='-1', command=self.doDown)
        cornerButton = ttk.Button(self, text='max positive', command=self.doCorner)
        self.modulusEntry = jtk.RTVEntry(self, self.modulusOK,
                                        self.master.doBad, width=11)
        cbValues = []
        i = 1
        while 1<<i <= self.theModel.MAXMODULUS:
            cbValues.append(i)
            i += 1
        self.bits = ttk.Combobox(self, values=cbValues,
                            state='readonly', width=3, takefocus=False)
        self.bits.bind("<<ComboboxSelected>>", self.doBits)
        self.modulusButton = ttk.Button(self, text="Set Modulus",
                            state="disabled", command=self.doModulus)
        
        upButton.grid(column=0, row=0)
        downButton.grid(column=1, row=0)
        cornerButton.grid(column=2, row=0)
        
        self.bits.grid(column=0, row=1)
        self.modulusEntry.grid(column=1, row=1)
        self.modulusButton.grid(column=2, row=1)
        
        ttk.Label(self, text="bits").grid(column=0, row=2)
        ttk.Label(self, text="proposed modulus").grid(column=1, row=2)

        self.bellButton = jtk.Checkbutton(self, True, text='dingdong?')
        self.bellButton.grid(column=2, row=2)
        
        root = self.master.master
        root.bind("<Up>", self.doUp)        ## self.master does not
        root.bind("<Down>", self.doDown)    ## work must be root

    def doUp(self, evt=None): ## keypress sends event; button does not
        if self.theModel.isGoodValue():
            val = (self.theModel.getValue() + 1)
            val = val % self.theModel.getModulus()
            self.theModel.setValue(val)
            self.master.updateViews(None)
        return "break"

    def doDown(self, evt=None): ## keypress sends event; button does not
        if self.theModel.isGoodValue():
            val = (self.theModel.getValue() - 1)
            val = val % self.theModel.getModulus()
            self.theModel.setValue(val)
            self.master.updateViews(None)
        return "break"
        
    def doCorner(self):
        m = self.theModel.getModulus()
        self.theModel.setValue((m-1)//2)
        self.master.updateViews(None)
        
## These values don't need to be validated, so we turn it off.
## This means that only keystrokes cause validation, and that
## is good to know in modulusOK() and its subroutines.
    def doBits(self, evt):
##        bits = self.bits.current() + 1
        bits = int(self.bits.get())
        self.modulusEntry["validate"] = "none"
        self.modulusEntry.delete(0, tkinter.END)
        self.modulusEntry.insert(0, str(1<<bits))
        self.modulusEntry["validate"] = "key"
        self.modulusButton.state(["!disabled"])
        
## We have several responsibilities:
    ## Accept or reject the proposed new value.
    ## If the new value is acceptable then
        ## Enable or disable the "set modulus" button appropriately.
        ## If the new value is a power of 2, change the value of 
        ## the "bits" Combobox appropriately else make it empty.
        ## figure the correct string to display in the RTVEntry
    def modulusOK(self, string, dummy):
        OK, result = self.decide(string)
        if OK:
            self.setButtonState(result)
            self.setBits(result)
            if result is not None: ## parameter string is not empty
                if result == 0: string = ''
                else: string = str(result)
        return OK, string

    def decide(self, string):
        if string == '':
            logging.debug("empty string for modulus")
            return True, None
        try:
            result = int(string, 10) ## can we parse as integer??
        except:
            logging.debug("bad number format for modulus")
            return False, None
        ## number format is ok; is value in range?
        logging.debug("got result {}".format(result))
        return 0 <= result <= self.theModel.MAXMODULUS, result

    def setButtonState(self, result):
        if result is not None and result >= 2:
            self.modulusButton.state(["!disabled"])
        else:  ## string is empty or result is in [0, 1]
            self.modulusButton.state(["disabled"])

    def setBits(self, result):
        if result is not None and result >= 2:
            L2 = math.log(result, 2)
            i = int(L2)
            if L2 == i:# L2 is an integer, so result is a power of 2
                self.bits.set(i)
            else:
                self.bits.set('')
        else:  ## string is empty or result < 2
            self.bits.set('')
            
    def doModulus(self):
        modulus = self.modulusEntry.get()
        modulus = int(modulus) ## guaranteed to be good
        assert 2 <= modulus <= self.theModel.MAXMODULUS
        self.theModel.setModulus(modulus)
    
if __name__ == "__main__":
    
    class App(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            MAXMODULUS = 1<<16
            m = self.myModel = model.Model(self, MAXMODULUS)
            m.setModulus(MAXMODULUS)
            m.setValue(None)
            bp = ButtonPanel(self, m)
            bp.pack(padx=75)
            m.pack()

        def updateViews(self, skip):
            pass

        def updateDigits(self):
            pass

        def doBad(self):
            self.bell()

    logging.basicConfig(level=logging.DEBUG)
    root = tkinter.Tk()
    root.title("testing ButtonPanel")
    a = App(root)
    a.pack()
    root.mainloop()
