import tkinter
from tkinter import ttk
from anycache import CacheApp

"""
Display a small window that lets the user choose properties of a cache. The properties are:
blocksize, cashesize, and associativity (blocks per set).
The rules are:
    each of these numbers is a power of 2
    blocksize * associativity <= 8 (i.e. number of words per set <= 8 so display is not too wide)
    cachesize <= 16 * blocksize * associativity (i.e. number of sets <= 16 so display is not too long)
These rules are enforced by making the Comboboxes readonly (disallowing typing in them) and by
using postcommand options in the Combobox constructors to restrict the choices available.
Each post method is called after the user clicks its Combobox and before the choices appear.
In between, the post method computes what choices should be available according to the rules
and configures the Combobox to display only those.
If blocksize and associativity are large, then cachesize can be chosen large. If blocksize and/or
associativity are later made smaller, then cachesize must be made smaller also.
"""

class MainWindow(ttk.Frame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cacheSize = ttk.Combobox(self, width=5, state='readonly', postcommand=self.cachePost)
        self.cacheSize.set("16")
        self.blockSize = ttk.Combobox(self, width=3, state='readonly', postcommand=self.blockPost)
        self.blockSize.set("1")
        self.blockSize.bind("<<ComboboxSelected>>", self.fixCacheSize)
        self.associativity = ttk.Combobox(self, width=3, state='readonly', postcommand=self.assocPost)
        self.associativity.set("1")
        self.associativity.bind("<<ComboboxSelected>>", self.fixCacheSize)
        self.button = ttk.Button(self, text="Create new cache", command=self.create_window)
        
        ttk.Label(self, text="Block Size").grid(row=0, column=0, padx=20, pady=5)
        self.blockSize.grid(row=0, column=1)
        ttk.Label(self, text="Associativity").grid(row=1, column=0, padx=20, pady=5)
        self.associativity.grid(row=1, column=1)
        ttk.Label(self, text="Cache Size").grid(row=2, column=0, padx=20, pady=5)
        self.cacheSize.grid(row=2, column=1)
        self.button.grid(row=3, column=0, columnspan=2, padx=20, pady=5)
        self.app = None

    def create_window(self):
        t = tkinter.Toplevel(self)
        cs = self.cacheSize.get()
        cs = int(cs)
        bs = self.blockSize.get()
        bs = int(bs)
        ass = self.associativity.get()
        ass = int(ass)
##        if self.app is not None:        ## So far there is no limit to the number of cache windows
##            self.app.destroy()          ## that can be open at once. We'll see how that goes.
        self.app = CacheApp(t, 1024, cs, bs, ass)
        self.app.pack(expand=True, fill="both")

    def fixCacheSize(self, evt):        ## be sure that cacheSize stays within its range
        bs = self.blockSize.get()
        bs = int(bs)
        ass = self.associativity.get()
        ass = int(ass)
        limit = 16 * bs * ass
        cs = self.cacheSize.get()
        cs = int(cs)
        if cs > limit:
            self.cacheSize.set(limit)

    def blockPost(self):
        assoc = self.associativity.get()
        assoc = int(assoc)
        values = [1]
        value = 2
        while value * assoc <= 8:
            values.append(value)
            value *= 2
        self.blockSize['values'] = values
        

    def assocPost(self):
        block = self.blockSize.get()
        block = int(block)
        values = [1]
        value = 2
        while value * block <= 8:
            values.append(value)
            value *= 2
        self.associativity['values'] = values

    def cachePost(self):
        assoc = self.associativity.get()
        assoc = int(assoc)
        block = self.blockSize.get()
        block = int(block)
        values = [1]
        value = 2
        while value <= 16 * block * assoc:
            values.append(value)
            value *= 2
        self.cacheSize['values'] = values
       
if __name__ == "__main__":
    root = tkinter.Tk()
    root.title("Make a Cache")
    main = MainWindow(root)
    main.pack(side="left", fill="y", expand=True)
    root.mainloop()
