import tkinter
from tkinter import ttk
import logging

"""
Demonstrate that we can move an RTVEntry and
its text on top of a fat line in a canvas.
By using Frame.after(milliseconds, function)
to schedule the moves, we keep the Button
alive at all times, which would not be the
case if we used sleep to time the moves.
The width parameter to create_window() matters.
The width parameter to ttk.RTVEntry() seems not to.

Formula for creating lines in the canvas for the crawlers to follow:
    create ttk features (RTVEntry, Label, Combobox, etc.) to be the crawling objects
    pass the canvas, a line, and a feature to the Crawler constructor for each Crawler object needed.

    The move function below is a good example of how to make Crawler objects crawl.
"""


class Crawler:
    DIM = 50                    ## default height and width of the crawling feature and width of fat lines
    stepSize = 2
    stepInterval = 3
    going = False

    def defaultSize():                                      ## Class method; no "self" parameter
        return Crawler.DIM                                  ## called as Crawler.defaultSize()

    def __init__(self, canvas, lineid, feature, done=None):
        self.canvas, self.lineid, self.feature, self.done = canvas, lineid, feature, done
        self.dim = int(round(float(canvas.itemcget(lineid, 'width')))) # use line width for crawling box's height and width
        coordinateList = canvas.coords(lineid)                ## [x0, y0, x1, y1, ...]
        self.here = 0                                          ## index in coordinateList of x-coordinate of current segment
        self.details =     {"e" : Crawler.CrawlDetails( 1,  0, self.oneStepEast),
                            "w" : Crawler.CrawlDetails(-1,  0, self.oneStepWest),
                            "n" : Crawler.CrawlDetails( 0, -1, self.oneStepNorth),
                            "s" : Crawler.CrawlDetails( 0,  1, self.oneStepSouth)}
        self.initialx, self.initialy = self.initialAdjust(coordinateList[0], coordinateList[1], self.direction(), self.dim)
        self.featureid = self.canvas.create_window(self.initialx, self.initialy, window=feature, height=self.dim,
                              width=self.dim, anchor="c", tags="goes")    ## tags parameter left in as an example for later
        self.feature.state(["readonly"])
#        logging.debug("lineid, featureid: ", self.lineid, self.featureid)

        ############################## PUBLIC METHODS ########################################

    def reset(self, newValue=None):
        self.here = 0                                                       ## reset coordinate list to beginning
        self.canvas.coords(self.featureid, self.initialx, self.initialy)    ## move feature back to its starting point
        if newValue is not None:                                            ## take on new value if supplied
            self.set(newValue)

    def setVisible(self, willBeVisible):
        self.feature.setVisible(willBeVisible)

    def set(self, newValue):
        self.feature.state(["!readonly"])
        if hasattr(self.feature, "set"):
            self.feature.set(newValue)
        elif hasattr(self.feature, "delete"):                               ## ttk.RTVEntry
            self.feature.delete(0, "end")
            self.feature.insert(0, newValue)
        else:                                                               ## ttk.Label
            self.feature.configure(text=newValue)
        self.feature.state(["readonly"])

    def get(self):
        return self.feature.get()
    
    def start(self, done="default"):                ## if supplied, done will replace default done method from constructor
        # print("starting: ", self.featureid)
        if done != "default":                       ## allow None to be specified here
            self.done = done
        self.canvas.after(0, self.move)             ## schedule the first step

    def stop(self):
        Crawler.going = False
        print(Crawler.going)

        ############################## PRIVATE METHODS ########################################
    
## self.canvas.coords(self.featureid) returns the coordinates of the crawling feature.
## They are actually the coordinates of the center point of the feature.
## Fat line coordinates are those of the endpoints of a skinny line down the center of the fat line.
## When fat line turns 90 degrees, the old segment appears to be extended by half the width of the new segment,
## so the leading edge of the crawling feature is exactly at the far edge of the fat line when its center
## reaches the end of the skinny line. However, the last segment is not extended in this way by the width
## of a next line that is not there. Therefor, we have to stop sooner by half the size (width/height)
## of the feature when we get to the end of the last segment. All this works fine as long as the size
## of the feature matches the thickness of the fat lines.

    def oneStepEast(self, stepSize):
#        print("moving ", self.featureid, " east")
        self.canvas.move(self.featureid, stepSize, 0)             ## take one step to the right
        x, y = self.canvas.coords(self.featureid)                 ## find where we are now
        stop = self.canvas.coords(self.lineid)[self.here + 2]     ## x-coordinate of far (right) end of line
        if self.isLastSegment():                                   ## no next line; correct for its width not being there
            stop -= self.dim / 2
        return x < stop                                            ## True if more in this segment             

    def oneStepSouth(self,stepSize):
        self.canvas.move(self.featureid, 0, stepSize)             ## take one step down
        x, y = self.canvas.coords(self.featureid)                 ## find where we are now
        stop = self.canvas.coords(self.lineid)[self.here + 3]     ## y-coordinate of far (bottom) end of line
        if self.isLastSegment():
            stop -= self.dim / 2
        return y < stop                                             ## True if more in this segment             

    def oneStepWest(self, stepSize):
        self.canvas.move(self.featureid, -stepSize, 0)            ## take one step to the left
        x, y = self.canvas.coords(self.featureid)                 ## find where we are now
        stop = self.canvas.coords(self.lineid)[self.here + 2]     ## x-coordinate of far (left) end of line
        if self.isLastSegment():
            stop += self.dim / 2
        return x > stop                                             ## True if more in this segment             

    def oneStepNorth(self, stepSize):
        self.canvas.move(self.featureid, 0, -stepSize)            ## take one step up
        x, y = self.canvas.coords(self.featureid)                 ## find where we are now
        stop = self.canvas.coords(self.lineid)[self.here + 3]     ## y-coordinate of far (top) end of line
        if self.isLastSegment():
            stop += self.dim / 2
        return y > stop                                            ## True if more in this segment             

    def direction(self):                                           ## hide all the ugliness in one place
        coords = self.canvas.coords(self.lineid)
        here = self.here
        if coords[here] == coords[here + 2]:                      ## vertical line since x-coordinates are equal
            if coords[here + 1] < coords[here + 3]:               ## look at y-coordinates
                return "s"
            elif coords[here + 1] > coords[here + 3]:
                return "n"
            else:
                assert False, "line segment with same x-coords and same y-coords"
        elif coords[here + 1] == coords[here + 3]:                ## horizontal line since y-coordinates are equal
            if coords[here] < coords[here + 2]:                   ## look at x-coordinates
                return "e"
            elif coords[here] > coords[here + 2]:
                return "w"
            else:
                assert False, "line segment with same x-coords and same y-coords"
        else:
             assert False, "line segment must be horizontal or vertical "

    def isLastSegment(self):
        return self.here >= len(self.canvas.coords(self.lineid)) - 4

## We have to correct the starting position of the center point of the feature to allow for its size.
## Cf. long comment above the oneStepXXX() methods.
    def initialAdjust(self, initialx, initialy, initialdirection, dimension):
        x = initialx + self.details[initialdirection].xstep * dimension / 2       ## one of these won't change
        y = initialy + self.details[initialdirection].ystep * dimension / 2       ## because its step value is 0
        return x, y

    def move(self):
        if Crawler.going:
            # print("moving: ", self.featureid)
            more = self.oneStep(self.stepSize)                    ## more == true iff this step did not reach end of line
            if more:
                self.canvas.after(self.stepInterval, self.move)   ## schedule another move
            else:
                logging.debug("Crawler {} got done = {}".format(self.lineid, self.done))
                if self.done is not None:
                    self.canvas.after(0, self.done)                   ## schedule completion method
        else:
            self.canvas.after(50, self.move)  ## try again later

    def oneStep(self, stepSize):
        # print("one step for: ", self.featureid)
        if self.here <= len(self.canvas.coords(self.lineid)) - 4:     ## more segments to crawl
            moreInThisSegment = self.oneStepInSegment(stepSize)
            if not moreInThisSegment:
                self.here += 2                                      ## go to next segment of line
            return True
        else:
            return False                                            ## we have reached the end

    def oneStepInSegment(self, stepSize):
        dir = self.direction()
        thismove = self.details[dir].moveFunction
        return thismove(stepSize)                                  ## True if more in this segment

    class CrawlDetails:                                            ## inner class in Crawler class
        def __init__(self, xstep, ystep, moveFunction):
            self.xstep, self.ystep, self.moveFunction = xstep, ystep, moveFunction

if __name__ == "__main__":

    class App(ttk.Frame):
        def __init__(self, parent):
            super().__init__(parent)
            self.canv = tkinter.Canvas(self, width=600, height=400)
            self.canv.pack()
            ## feature can be an RTVEntry or a Combobox or prolly one of several other things
            ## its dimensions will be determined by parameters to Canvas.create_window in Crawler.__init__()
            feature = ttk.Combobox(self, values=[1,2,3,4,5], takefocus=False)                   ## feature is a Combobox
            line1 = self.canv.create_line(250, 50, 500, 50, 500, 250, 250, 250, 250, 100,
                                width=Crawler.defaultSize(), joinstyle="miter", arrow=tkinter.LAST)   # arrowhead at end of line
            self.crlOne = Crawler(self.canv, line1, feature)
            self.crlOne.set("1")

            feature = ttk.Label(self)
            line2 = self.canv.create_line(50, 350, 550, 350,
                                width=Crawler.defaultSize(), joinstyle="miter", arrow=tkinter.LAST)
            self.crlTwo = Crawler(self.canv, line2, feature)
            self.crlTwo.set("xxx")
            
            self.button = ttk.Button(self, text="start", command=self.pressed)
            self.button.pack()
            self.crlOne.start(self.end1)             ## call end1 on completion

        def end1(self):
            self.crlTwo.start(self.reset1)           ## reset1 on completion

        def reset1(self):
            self.crlOne.reset("ab")
            self.crlTwo.reset("cd")
            self.crlOne.start(self.reset2)           ## reset2 on completion
            self.crlTwo.start(self.reset2)
            self.joinCount = 0

        def reset2(self):
            self.joinCount += 1
            if self.joinCount == 2:                   # we are second one here; reset everything for restart
                self.crlOne.reset("1")
                self.crlTwo.reset("xxx")
                self.button["text"] = "start"
                Crawler.going = False                 # but don't actually start
                self.crlOne.start(self.end1)

        def pressed(self):                              ## start/stop button handler
            Crawler.going = not Crawler.going
            if Crawler.going:
                self.button["text"] = "stop"
            else:
                self.button["text"] = "start"
 
    def main():
        root = tkinter.Tk()
        root.title("Moving Items")
        a = App(root)
        a.pack()
        root.mainloop()

    main()
