'''
Demonstrates stackframes.
Selecting a framesize decrements the stack pointer appropriately
and gives the words in the new stackframe a contrasting color.
Pushing the pop button removes the top stackframe.
'''

import tkinter
from tkinter import ttk
import jtk
import random
import math

STACKBOTTOM = 0x7ffff018
OLDSTACKSIZE = 8
INITIALPOINTER = STACKBOTTOM - 4*(OLDSTACKSIZE-1)   ## initial SP is at bottom of old stack

class StackPointer(ttk.Frame):
    def __init__(self, parent, mem):
        super().__init__(parent)
        self.memory = mem
        self.history = [OLDSTACKSIZE - 1]           ## index of initial SP
        self.value = 0
        
        self.display = ttk.Entry(self,  width=16)
        self.setValue(INITIALPOINTER)               ## leaves display in readonly state
        self.cbox = cbox = ttk.Combobox(self, values=['1', '2', '3', '4', '5'],
                                    width=3, state='readonly', takefocus=False)
        cbox.bind("<<ComboboxSelected>>", self.doPush)
        cbox.set("2")
        self.pb = ttk.Button(self, text="Pop Frame", command=self.doPop, state="disabled")

        ttk.Label(self, text="Stack Pointer").grid(row=0, column=0)        
        self.display.grid(row=1, column=0)
        ttk.Label(self, text="").grid(row=2, column=0)        
        ttk.Label(self, text="New Frame Size").grid(row=3, column=0)
        cbox.grid(row=4, column=0, padx=10)
        ttk.Label(self, text="").grid(row=5, column=0)        
        self.pb.grid(row=6, column=0)

    def setValue(self, newValue):
        self.value = newValue
        valString = jtk.dividedHexStr(newValue)
        self.display.state(["!readonly"])
        self.display.delete(0, "end")
        self.display.insert(0, valString)
        self.display.state(["readonly"])

    def doPush(self, evt=None):
        op = self.cbox.get()        ## number of words to add to stack
        op = int(op)
        newSp = self.value - 4*op
        self.setValue(newSp)        ## new value of stack pointer
        idx = self.memory.indexFromAddr(newSp)
        self.history.append(idx)    ## history is a list of indexes of successive stack pointers
##        print(self.history, self.topRng())
        colorSeq = len(self.history) - 1         ## get next color
        for i in self.topRng():     ## color each word in new stackframe
            id = self.memory.idFromIndex(i)
            self.memory.tree.item(id, tags=jtk.DISTINCT_COLORS[colorSeq])
        self.pb.state(["!disabled"])            ## pop is now ok
        self.memory.doBaseDisplacements(newSp)

    def topRng(self):                       ## range of indices in top stackframe
        hlen = len(self.history)
        currSp = self.history[hlen - 1]     ## index of current SP is at end of history list
        prevSp = self.history[hlen - 2]     ## index of previous SP is just before it
        return range(prevSp+1, currSp+1)

    def doPop(self, evt=None):
        hlen = len(self.history)
        currSp = self.history[hlen - 1]     ## index of current SP is at end of history list
        prevSp = self.history[hlen - 2]     ## index of previous SP is just before it
##        print(self.history, currSp, prevSp)
        currSp = self.memory.addrFromIndex(currSp)     ## the actual address in current SP
        prevSp = self.memory.addrFromIndex(prevSp)     ## the actual address in previous SP
        self.setValue(prevSp)
        for i in self.topRng():             ## uncolor and clear old stackframe
            id = self.memory.idFromIndex(i)
            self.memory.tree.item(id, tags="white")
            self.memory.tree.set(id, "base-displacement", "")
        self.history = self.history[:-1]    ## remove current SP from history list
        if hlen <= 2:                       ## disable pop button if no more frames to pop
            self.pb.state(["disabled"])
        self.memory.doBaseDisplacements(prevSp)

class Memory(jtk.McListBox):
    def __init__(self, parent, headers, size):
        super().__init__(parent, headers, selectmode="none")
        self.size = size
        for idx in range(size):
            addr = self.addrFromIndex(idx)
            self.tree.insert('', 'end', self.idFromIndex(idx),
                     values=(jtk.dividedHexStr(addr), idx))
            id = self.idFromIndex(idx)
            if addr >= INITIALPOINTER:
                self.tree.item(self.idFromIndex(idx), tags="red")
        self.tree["height"] = size
        for i in range(len(jtk.DISTINCT_COLORS)):
            self.tree.tag_configure(jtk.DISTINCT_COLORS[i], background=jtk.DISTINCT_COLORS[i])
        self.tree.tag_configure("red", background="red")
        self.tree.tag_configure("white", background="white")
        self.doBaseDisplacements(INITIALPOINTER)

    def addrFromIndex(self, idx):
        return STACKBOTTOM - 4*idx

    def indexFromAddr(self, addr):
        return (STACKBOTTOM - addr) // 4

    def doBaseDisplacements(self, lowerLimit):
        addr = STACKBOTTOM
        while addr >= lowerLimit:
            idx = self.indexFromAddr(addr)
            id = self.idFromIndex(idx)
            displacement = addr - lowerLimit
            self.tree.set(id, "base-displacement", "{}($sp)".format(displacement))
            addr -= 4            

class App(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        self.mem = Memory(self, ("     address", "  contents  ", "base-displacement"), 32)
        self.sp = StackPointer(self, self.mem)
        self.sp.pack(side="left", padx=10)
        self.mem.pack(side="left", padx=15)
    
root = tkinter.Tk()
root.title("Stackframes")
app = App(root)
app.pack()
root.mainloop()
