import tkinter
from tkinter import ttk
import model
import jtk
import logging
import math

class ValuePanel(ttk.Frame):
    def __init__(self, parent, index, baseValues, myModel, rownum=0, width=33):
        super().__init__(parent, padding="3 3 3 3")
        self.myModel = myModel
        ## enable/disable is controlled by onBase()
        self.value = jtk.RTVEntry(self, self.resultOK, self.master.doBad,
                        state='disabled', width=width)
        
        ## Checkbutton controls whether this view is signed or unsigned      
        self.signButton = jtk.Checkbutton(self, False, takefocus=False,
                            command=self.onSign, onvalue=True, offvalue=False)
        
        ## Combobox selects the base for this view
        self.base = ttk.Combobox(self, values=baseValues,
                            state='readonly', width=3, takefocus=False)
        self.base.bind("<<ComboboxSelected>>", self.onBase)

        self.signButton.grid(column=0, row=rownum)
        self.base.grid(column=1, row=rownum)
        self.value.grid(column=2, row=rownum, padx = 10)
        self.index = index       ## We expect to be in a list of ValuePanels

## acceptance criterion for typed strings
## return value MUST be (Boolean, something)
## if the Boolean is True, something MUST be a string
    def resultOK(self, string, dummy):
        isSigned = self.signButton.isSelected()
        if string == '' or isSigned and string in ['-', '+']:
            self.myModel.setValue(None)
            self.master.updateViews(self)
            return True, string        
        base = self.getBase()
        assert base is not None ## shouldn't be None; we got a keystroke
        modulus = self.myModel.getModulus()
        OK, result = self.decide(string, base, modulus, isSigned)
        if OK:
            if result < 0:
                result += modulus ## model stores ONLY positive values
            self.myModel.setValue(result)
            self.master.updateViews(self)
            ## interesting; typed-in view gets updated last
            return True, string.upper() ## display upper even if lower was typed
        else:
            return False, None

    def decide(self, string, base, modulus, signed):
        logging.debug("deciding on '{}'".format(string))
        try:
            result = int(string, base) ## can we parse as integer??
        except:
            logging.debug("bad number format")
            return False, None
        else:  ## number format is ok; is value in range?
            logging.debug("got result {}".format(result))
            isSigned = self.signButton.isSelected()
            maximum = self.myModel.maxPossible(isSigned)
            minimum = self.myModel.minPossible(isSigned)
            logging.debug("min max are {} {}".format(maximum, minimum))
            if minimum <= result <= maximum:
                logging.debug("liking it")
                return True, result
            else:
                logging.debug("not liking it")
                return False, result
                    
    def onSign(self):
        self.setDigits()
        if self.myModel.isGoodValue():
            self.setNum(self.myModel.getValue())     ## direct change
            
## keep up with how many digits we display in this base
## so we can pad with leading zeros
    def setDigits(self):
        base = self.getBase()
        if base is not None:
            isSigned = self.signButton.isSelected()
            m = self.myModel.maxPossible(isSigned)
            s = jtk.radStr(m, base)
            logging.debug('setting digits to {}'.format(len(s)))
            self.digits = len(s) 

    def onBase(self, event):
        if self.getBase() is not None:
            self.value.state(["!disabled"])
            self.setDigits()
            if self.myModel.isGoodValue():
                self.setNum(self.myModel.getValue()) ## direct change
        else:
            self.setNum(None)                        ## direct change
            self.value.state(["disabled"])
        dis = self.value.instate(["disabled"])
        logging.info("onBase selected {} in panel {} disabled is {}"
                           .format(self.getBase(), self.index, dis))
            
    def getBase(self):  ## returns None or integer
        cur = self.base.current()
        if cur == 0:
            return None
        else:
            return int(self.base.get()) ## allow non-consecutive values
        
    def getValue(self):  ## returns  integer
        assert self.getBase() is not None
        return int(self.value.get(), self.getBase())

    def setNum(self, newValue):  ## newValue is integer or None
    ## It's safe to use unvalidatedSetString here because the new value is the empty
    ## string or it has been fetched from the Model and is therefore known to be valid.
        if newValue is None:
            self.value.unvalidatedSetStr('')
        else:
            bv = self.getBase()
            if bv is not None:
                isSigned = self.signButton.isSelected()
                modulus = self.myModel.getModulus()
                if isSigned and newValue > self.myModel.maxPossible(True):
                    s = jtk.radStr(newValue-modulus, bv, self.digits)
                else:
                    s = jtk.radStr(newValue, bv, self.digits)
                self.value.unvalidatedSetStr(s)        

class TopValuePanel(ValuePanel):
    def __init__(self, parent, baseValues, myModel, width=33):
        ## stuff goes in row 1 to leave room for header in row 0
        super().__init__(parent, 1, baseValues, myModel, 1)
        ttk.Label(self, text='\u00B1?').grid(column=0, row=0)
        ttk.Label(self, text='base').grid(column=1, row=0)
        ttk.Label(self, text='value').grid(column=2, row=0)
            
SHORT_BASES = False

class App(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)
        logging.basicConfig(level=logging.INFO)
        self.myModel = myModel = model.Model(self, 1024)
        if SHORT_BASES:
            bases = ['', 2, 10, 16]
        else:
            bases = ['']
            for i in range(2,37): ## gives us 2-36
                bases.append(i)
        self.vp = vp = ValuePanel(self, 0, bases, myModel)
        vp.pack(padx=10)
        ttk.Separator(self, orient="horizontal")
        myModel.pack()
        myModel.setValue(None)
        root.bind("<Up>", self.doUp)
        root.bind("<Down>", self.doDown)

    def doUp(self, evt=None): ## keypress sends event; button does not
        if self.myModel.isGoodValue():
            val = (self.myModel.getValue() + 1)
            val = val % self.myModel.getModulus()
            self.myModel.setValue(val)
            self.updateViews(None)
            return "break"

    def doDown(self, evt=None): ## keypress sends event; button does not
        if self.myModel.isGoodValue():
            val = (self.myModel.getValue() - 1)
            val = val % self.myModel.getModulus()
            self.myModel.setValue(val)
            self.updateViews(None)
            return "break"

    def updateViews(self, skip):
        if self.vp != skip:
            self.vp.setNum(self.myModel.getValue())

    def doBad(self):
        self.bell()

if __name__ == "__main__":
    root = tkinter.Tk()
    root.title("testing valuepanel")
    a = App(root)
    a.pack()
    root.mainloop()
